package org.kamiblue.client.module.modules.render

import net.minecraft.client.renderer.GlStateManager
import net.minecraft.client.shader.Shader
import net.minecraft.entity.Entity
import net.minecraft.entity.item.EntityItem
import net.minecraft.entity.item.EntityXPOrb
import net.minecraft.entity.projectile.EntityArrow
import net.minecraft.entity.projectile.EntityThrowable
import net.minecraft.util.ResourceLocation
import net.minecraftforge.client.event.EntityViewRenderEvent
import net.minecraftforge.fml.common.gameevent.TickEvent
import org.kamiblue.client.event.Phase
import org.kamiblue.client.event.SafeClientEvent
import org.kamiblue.client.event.events.RenderEntityEvent
import org.kamiblue.client.event.events.RenderWorldEvent
import org.kamiblue.client.mixin.extension.*
import org.kamiblue.client.module.Category
import org.kamiblue.client.module.Module
import org.kamiblue.client.util.*
import org.kamiblue.client.util.EntityUtils.getTargetList
import org.kamiblue.client.util.color.ColorHolder
import org.kamiblue.client.util.graphics.ESPRenderer
import org.kamiblue.client.util.graphics.GlStateUtils
import org.kamiblue.client.util.graphics.KamiTessellator
import org.kamiblue.client.util.graphics.ShaderHelper
import org.kamiblue.client.util.threads.runSafe
import org.kamiblue.client.util.threads.safeListener
import org.kamiblue.event.listener.listener
import org.lwjgl.opengl.GL11.GL_MODELVIEW
import org.lwjgl.opengl.GL11.GL_PROJECTION

internal object ESP : Module(
    name = "ESP",
    category = Category.RENDER,
    description = "Highlights entities"
) {
    private val page = setting("Page", Page.ENTITY_TYPE)

    /* Entity type settings */
    private val all0 = setting("All Entities", false, page.atValue(Page.ENTITY_TYPE))
    private val all by all0
    private val experience by setting("Experience", false, page.atValue(Page.ENTITY_TYPE) and all0.atFalse())
    private val arrows by setting("Arrows", false, page.atValue(Page.ENTITY_TYPE) and all0.atFalse())
    private val throwable by setting("Throwable", false, page.atValue(Page.ENTITY_TYPE) and all0.atFalse())
    private val items by setting("Items", true, page.atValue(Page.ENTITY_TYPE) and all0.atFalse())
    private val players0 = setting("Players", true, page.atValue(Page.ENTITY_TYPE) and all0.atFalse())
    private val players by players0
    private val friends by setting("Friends", false, page.atValue(Page.ENTITY_TYPE) and all0.atFalse() and players0.atTrue())
    private val sleeping by setting("Sleeping", false, page.atValue(Page.ENTITY_TYPE) and all0.atFalse() and players0.atTrue())
    private val mobs0 = setting("Mobs", true, page.atValue(Page.ENTITY_TYPE) and all0.atFalse())
    private val mobs by mobs0
    private val passive by setting("Passive Mobs", false, page.atValue(Page.ENTITY_TYPE) and all0.atFalse() and mobs0.atTrue())
    private val neutral by setting("Neutral Mobs", true, page.atValue(Page.ENTITY_TYPE) and all0.atFalse() and mobs0.atTrue())
    private val hostile by setting("Hostile Mobs", true, page.atValue(Page.ENTITY_TYPE) and all0.atFalse() and mobs0.atTrue())
    private val invisible by setting("Invisible", true, page.atValue(Page.ENTITY_TYPE) and all0.atFalse())
    private val range by setting("Range", 32.0f, 8.0f..64.0f, 0.5f, page.atValue(Page.ENTITY_TYPE))

    /* Rendering settings */
    private val mode = setting("Mode", Mode.SHADER, page.atValue(Page.RENDERING))
    private val hideOriginal by setting("Hide Original", false, page.atValue(Page.RENDERING) and mode.atValue(Mode.SHADER))
    private val filled by setting("Filled", false, page.atValue(Page.RENDERING) and mode.atValue(Mode.BOX, Mode.SHADER))
    private val outline by setting("Outline", true, page.atValue(Page.RENDERING) and mode.atValue(Mode.BOX, Mode.SHADER))
    private val r by setting("Red", 155, 0..255, 1, page.atValue(Page.RENDERING) and mode.atValue(Mode.BOX, Mode.SHADER))
    private val g by setting("Green", 144, 0..255, 1, page.atValue(Page.RENDERING) and mode.atValue(Mode.BOX, Mode.SHADER))
    private val b by setting("Blue", 255, 0..255, 1, page.atValue(Page.RENDERING) and mode.atValue(Mode.BOX, Mode.SHADER))
    private val aFilled by setting("Filled Alpha", 63, 0..255, 1, page.atValue(Page.RENDERING) and mode.atValue(Mode.BOX, Mode.SHADER))
    private val aOutline by setting("Outline Alpha", 255, 0..255, 1, page.atValue(Page.RENDERING) and mode.atValue(Mode.BOX, Mode.SHADER))
    private val blurRadius by setting("Blur Radius", 0f, 0f..16f, 0.5f, page.atValue(Page.RENDERING) and mode.atValue(Mode.SHADER))
    private val width by setting("Width", 2f, 1f..8f, 0.25f, page.atValue(Page.RENDERING))

    private enum class Page {
        ENTITY_TYPE, RENDERING
    }

    private enum class Mode {
        BOX, GLOW, SHADER
    }

    private val entityList = LinkedHashSet<Entity>()

    private val shaderHelper = ShaderHelper(ResourceLocation("shaders/post/esp_outline.json"), "final")
    private val frameBuffer = shaderHelper.getFrameBuffer("final")

    init {
        listener<RenderEntityEvent.All> {
            if (it.phase == Phase.PRE
                && mode.value == Mode.SHADER
                && !mc.renderManager.renderOutlines
                && hideOriginal
                && entityList.contains(it.entity)) {
                it.cancel()
            }
        }

        safeListener<EntityViewRenderEvent.FogColors> { event ->
            shaderHelper.shader?.listFrameBuffers?.forEach {
                it.setFramebufferColor(event.red, event.green, event.blue, 0.0f)
            }
        }

        safeListener<RenderWorldEvent>(69420) {
            when (mode.value) {
                Mode.BOX -> {
                    val color = ColorHolder(r, g, b)
                    val renderer = ESPRenderer()
                    renderer.aFilled = if (filled) aFilled else 0
                    renderer.aOutline = if (outline) aOutline else 0
                    renderer.thickness = width
                    for (entity in entityList) {
                        renderer.add(entity, color)
                    }
                    renderer.render(true)
                }
                Mode.SHADER -> {
                    drawEntities()
                    drawShader()
                }
                else -> {
                    // Glow Mode
                }
            }
        }
    }

    private fun drawEntities() {
        // Clean up the frame buffer and bind it
        frameBuffer?.framebufferClear()
        frameBuffer?.bindFramebuffer(false)

        val prevRenderOutlines = mc.renderManager.renderOutlines

        GlStateUtils.texture2d(true)
        GlStateUtils.cull(true)

        // Draw the entities into the framebuffer
        for (entity in entityList) {
            val renderer = mc.renderManager.getEntityRenderObject<Entity>(entity) ?: continue

            val partialTicks = KamiTessellator.pTicks()
            val yaw = entity.prevRotationYaw + (entity.rotationYaw - entity.prevRotationYaw) * partialTicks
            val pos = EntityUtils.getInterpolatedPos(entity, partialTicks)
                .subtract(mc.renderManager.renderPosX, mc.renderManager.renderPosY, mc.renderManager.renderPosZ)

            renderer.setRenderOutlines(true)
            renderer.doRender(entity, pos.x, pos.y, pos.z, yaw, partialTicks)
            renderer.setRenderOutlines(prevRenderOutlines)
        }

        GlStateUtils.texture2d(false)
    }

    private fun drawShader() {
        // Push matrix
        GlStateManager.matrixMode(GL_PROJECTION)
        GlStateManager.pushMatrix()
        GlStateManager.matrixMode(GL_MODELVIEW)
        GlStateManager.pushMatrix()

        shaderHelper.shader?.render(KamiTessellator.pTicks())

        // Re-enable blend because shader rendering will disable it at the end
        GlStateUtils.blend(true)
        GlStateUtils.depth(false)

        // Draw it on the main frame buffer
        mc.framebuffer.bindFramebuffer(false)
        frameBuffer?.framebufferRenderExt(mc.displayWidth, mc.displayHeight, false)

        // Revert states
        GlStateUtils.blend(true)
        GlStateUtils.depth(true)
        GlStateUtils.texture2d(false)
        GlStateManager.depthMask(false)
        GlStateUtils.cull(false)

        // Revert matrix
        GlStateManager.matrixMode(GL_PROJECTION)
        GlStateManager.popMatrix()
        GlStateManager.matrixMode(GL_MODELVIEW)
        GlStateManager.popMatrix()
    }

    init {
        safeListener<TickEvent.ClientTickEvent> {
            entityList.clear()
            entityList.addAll(getEntityList())

            when (mode.value) {
                Mode.GLOW -> {
                    if (entityList.isNotEmpty()) {
                        for (shader in mc.renderGlobal.entityOutlineShader.listShaders) {
                            shader.shaderManager.getShaderUniform("Radius")?.set(width)
                        }

                        for (entity in world.loadedEntityList) { // Set glow for entities in the list. Remove glow for entities not in the list
                            entity.isGlowing = entityList.contains(entity)
                        }
                    } else {
                        resetGlow()
                    }
                }
                Mode.SHADER -> {
                    shaderHelper.shader?.let {
                        for (shader in it.listShaders) {
                            setShaderSettings(shader)
                        }
                    }
                }
                else -> {
                    // Box Mode
                }
            }
        }
    }

    private fun SafeClientEvent.getEntityList(): List<Entity> {
        val playerSettings = arrayOf(players, friends, sleeping)
        val mob = arrayOf(mobs, passive, neutral, hostile)
        val entityList = ArrayList<Entity>()
        if (all) {
            for (entity in world.loadedEntityList) {
                if (entity == mc.renderViewEntity) continue
                if (player.getDistance(entity) > range) continue
                entityList.add(entity)
            }
        } else {
            entityList.addAll(getTargetList(playerSettings, mob, invisible, range, ignoreSelf = false))
            for (entity in world.loadedEntityList) {
                if (entity == player) continue
                if (player.getDistance(entity) > range) continue
                if (entity is EntityXPOrb && experience
                    || entity is EntityArrow && arrows
                    || entity is EntityThrowable && throwable
                    || entity is EntityItem && items) {
                    entityList.add(entity)
                }
            }
        }
        return entityList
    }

    private fun setShaderSettings(shader: Shader) {
        shader.shaderManager.getShaderUniform("color")?.set(r / 255f, g / 255f, b / 255f)
        shader.shaderManager.getShaderUniform("outlineAlpha")?.set(if (outline) aOutline / 255f else 0f)
        shader.shaderManager.getShaderUniform("filledAlpha")?.set(if (filled) aFilled / 255f else 0f)
        shader.shaderManager.getShaderUniform("width")?.set(width)
        shader.shaderManager.getShaderUniform("Radius")?.set(blurRadius)
    }

    init {
        onDisable {
            resetGlow()
        }

        mode.listeners.add {
            resetGlow()
        }
    }

    private fun resetGlow() {
        runSafe {
            for (shader in mc.renderGlobal.entityOutlineShader.listShaders) {
                shader.shaderManager.getShaderUniform("Radius")?.set(2f) // default radius
            }

            for (entity in world.loadedEntityList) {
                entity.isGlowing = false
            }
        }
    }
}